#include <stdlib.h>

#include "StdAfx.h"
#include "PlatformDisplay.h"

const COLORREF WHITE = 0xFFFFFF;

enum ORIENTATION
{
	ORTHOGONAL = 0,
	DIAGONALTRUE = 1,
	DIAGONALFAST = 2,
	DIAGONALBAROQUE = 3
};

struct WindowsDisplayParams
{
	int cellSize;
	HDC dc;
	COLORREF color;
	ORIENTATION orientation;
	BOOL mosaic;
};

class WindowsDisplay : private WindowsDisplayParams
{
private:
	int boxSize;
	int xOffset, yOffset;
	int lastBoxSize;
	COLORREF lastColor;

	DWORD rop;
	HDC memDC;
	HBRUSH hBrush;
	HBITMAP hBitmap;
	HGDIOBJ hOldBrush;
	HGDIOBJ hOldBitmap;

private:
	BOOL usesBitmap(void) const
	{
		return orientation == DIAGONALTRUE && cellSize > 1;
	}

	void OpenBrush(void)
	{
		if (color != WHITE)
		{
			hBrush = CreateSolidBrush(color);
			hOldBrush = SelectObject(dc, hBrush);
		}

		rop = color == WHITE ? DSTINVERT : PATINVERT;
	}

	void CloseBrush(void)
	{
		if (color != WHITE)
		{
			SelectObject(dc, hOldBrush);
			DeleteObject(hBrush);
		}
	}

	void OpenBitmap(void)
	{
		if (boxSize != lastBoxSize)
		{
			if (hBitmap) DeleteObject(hBitmap);
			hBitmap = CreateCompatibleBitmap(dc, boxSize, boxSize);
		}

		memDC = CreateCompatibleDC(dc);
		hOldBitmap = SelectObject(memDC, hBitmap);

		if (boxSize != lastBoxSize || color != lastColor)
		{
			PatBlt(memDC, 0, 0, boxSize, boxSize, BLACKNESS);

			HBRUSH hBrush = CreateSolidBrush(color);
			HGDIOBJ hOldBrush = SelectObject(memDC, hBrush);

			for (int y = 0; y < boxSize; y++)
			{
				const int x = abs(cellSize - 1 - y);
				const int width = abs(boxSize - 2 * x);
				PatBlt(memDC, x, y, width, 1, PATCOPY);
			}

			SelectObject(memDC, hOldBrush);
			DeleteObject(hBrush);
		}

		lastBoxSize = boxSize;
		lastColor = color;

		rop = SRCINVERT;
	}

	void CloseBitmap(void)
	{
		SelectObject(memDC, hOldBitmap);
		DeleteObject(memDC);
	}

	void PlotOrthogonal(const int x, const int y)
	{
		PatBlt(dc,
			x * cellSize + xOffset,
			y * cellSize + yOffset,
			boxSize,
			boxSize,
			rop);
	}

	void PlotDiagonal(const int x, const int y)
	{
		if (usesBitmap())
		{
			BitBlt(dc,
				(x - y) * cellSize + xOffset,
				(x + y) * cellSize + yOffset,
				boxSize,
				boxSize,
				memDC,
				0,
				0,
				rop);
		}
		else
		{
			PatBlt(dc,
				(x - y) * cellSize + xOffset,
				(x + y) * cellSize + yOffset,
				boxSize,
				boxSize,
				rop);
		}
	}

public:
	WindowsDisplay(void)
	{
		hBitmap = 0;
		lastBoxSize = 0;
	}

	~WindowsDisplay()
	{
		if (hBitmap) DeleteObject(hBitmap);
		hBitmap = 0;
	}

	void Open(const WindowsDisplay *const pParams)
	{
		dc = pParams->dc;
		cellSize = pParams->cellSize;
		color = pParams->color;
		orientation = pParams->orientation;
		mosaic = pParams->mosaic;

		switch (orientation)
		{
		case ORTHOGONAL:
			xOffset = cellSize > 2 && mosaic;
			yOffset = cellSize > 2 && mosaic;
			boxSize = cellSize - xOffset;
			break;

		case DIAGONALTRUE:
		case DIAGONALBAROQUE:
			xOffset = 1 - cellSize;
			yOffset = 1;
			boxSize = 2 * cellSize - 1;
			break;

		case DIAGONALFAST:
			xOffset = - (cellSize / 2);
			yOffset = (cellSize + 1) / 2;
			boxSize = cellSize;
			break;

		}

		if (usesBitmap()) OpenBitmap(); else OpenBrush();
	}

	void Plot(int x, int y)
	{
		switch (orientation)
		{
		case ORTHOGONAL:
			PlotOrthogonal(x, y);
			break;

		case DIAGONALTRUE:
		case DIAGONALFAST:
		case DIAGONALBAROQUE:
			PlotDiagonal(x, y);
			break;
		}
	}

	void Close()
	{
		if (usesBitmap()) CloseBitmap(); else CloseBrush();
	}
};

typedef WindowsDisplay PlatformDisplay;

PlatformDisplay pd;

void PlatformDisplayOpen(const PDISPLAYPARAMS params)
{
	pd.Open((PlatformDisplay *) params);
}

void PlatformDisplayPlot(const int x, const int y)
{
	pd.Plot(x, y);
}

void PlatformDisplayClose(void)
{
	pd.Close();
}
